<?php

namespace App\Http\Controllers\Booking;

use App\Http\Controllers\Controller;
use App\Http\Traits\CommonTrait;
use App\Http\Traits\EditLogTrait;
use App\Http\Traits\PermissionsTrait;
use App\Models\Booking\BookGr;
use App\Models\Booking\DeliveryOrder;
use App\Models\Booking\DeliveryOrderDesc;
use App\Models\Booking\GrDesc;
use App\Models\Master\Party;
use Carbon\Carbon;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;

class DeliveryOrderController extends Controller
{
    use PermissionsTrait;
    use EditLogTrait;
    use CommonTrait;

    public function index(Request $request){
        $test = $this->is_read_permitted($request->user->id, 'DeliveryOrder');
        if(!$test){
            return response()->json(['Permission Denied For The Action'],401);
        }

        //FETCH AND SEND ALL
        $dos = DeliveryOrder::with([
                    'branch:id,branch_name,branch_code', 'consignor:id,party_name', 'consignee:id,party_name',
                    'acc_party.party:id,account_id',
                    'do_desc',
                    'do_desc.source:id,station_name', 'do_desc.destination:id,station_name', 'do_desc.material:id,material_name'
                    ])->where('fin_year',$request->current_year)->get();
        return response()->json($dos);
    }

    public function show(Request $request, $pk){
        // FETCH AND SEND
        $do = DeliveryOrder::with([
                'branch:id,branch_name,branch_code', 'consignor:id,party_name', 'consignee:id,party_name',
                'acc_party.party:id,account_id',
                'do_desc',
                'book_gr:id,gr_number,do_id',
                'do_desc.source:id,station_name', 'do_desc.destination:id,station_name', 'do_desc.material:id,material_name'
            ])->find($pk);
        
        
        // Transform do_desc here
        $transformedDoDesc = $do->do_desc->map(function($doDesc) {
            
            $balance = $this->find_remaining_balance_for_do_desc($doDesc->id);

            $doDesc['remaining_qty'] = $balance['remaining_qty'];

            return $doDesc;

        });

        $do->setRelation('do_desc', $transformedDoDesc);

        return response()->json(['do'=>$do]);
    }

    public function store(Request $request){

        $test = $this->is_new_permitted($request->user->id, 'DeliveryOrder');
        if(!$test){
            return response()->json(['Permission Denied For The Action'],401);
        }
        
        $validatedData = $request->validate([
            'branch' => 'required',
            'do_number' => 'required|unique:delivery_orders,do_number',
            'acc_party' => 'required'
        ], [
            'branch.required' => 'Required',
            'do_number.required' => 'Required',
            'do_number.unique' => 'Already Exist !',
            'acc_party.required' => 'required'
        ]);

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction();  
            

            // IN CASE THE GR TYPE IS 'PRINTED', WE NEED TO CHECK IF THE BOOK IS ACTUALLY ISSUED TO THE BRANCH OR NOT
            $max_number = $this->get_max_number_from_do_number($request->do_number);
            
            
            // SAVING A NEW GR
            $do = new DeliveryOrder();

            $do->branch_id = $request->branch['id'];
            $do->do_number = $request->do_number;
            $do->max_number = $max_number;
            $do->invoice_number = $request->invoice_number;
            $do->acc_party_id = $request->acc_party['id'];
            $do->from_date = $request->from_date;
            $do->to_date = $request->to_date;
            $do->is_locked = $request->is_locked;
            $do->order_no = $request->order_no;
            $do->po_no = $request->po_no;
            $do->acc_party_id = $request->acc_party['id'];
            $do->consignor_id = $request->consignor?$request->consignor['id']:null;
            $do->consignor_name = $request->consignor?$request->consignor['party_name']:null;
            $do->consignee_id = $request->consignee?$request->consignee['id']:null;
            $do->consignee_name = $request->consignee?$request->consignee['party_name']:null;
            $do->invc_val = $request->invc_val;
            $do->time_alloted = $request->time_alloted;

            $do->fin_year = $request->current_year;
            $do->created_by = $request->user->id;
            $do->save();


            $do_descs = $request->do_descs;
            foreach ($do_descs as $desc) {
                if( !$desc['source'] || !$desc['unit']   )
                continue;
                
                $doDtls = new DeliveryOrderDesc();
                $doDtls->do_number = $do->do_number;
                $doDtls->do_id = $do->id;
                $doDtls->unit = $desc['unit'];
                $doDtls->material_id = $desc['material'] ? $desc['material']['id'] : null;
                $doDtls->rate = $desc['rate'];
                $doDtls->hire_rate = $desc['hire_rate'];
                $doDtls->loading_exp = $desc['loading_exp'];
                $doDtls->qty = $desc['qty'];
                $doDtls->source_id = $desc['source']['id'];
                $doDtls->source_name = $desc['source']['station_name'];
                $doDtls->destination_id =  $desc['destination'] ? $desc['destination']['id']: null;
                $doDtls->destination_name =  $desc['destination'] ? $desc['destination']['station_name']: null;
                $doDtls->eway_no = $desc['eway_no'];
                $doDtls->eway_dt = $desc['eway_dt'];
                $doDtls->item_desc = $desc['item_desc'];
                $doDtls->fin_year = $request->current_year;
                $doDtls->save();
            }

            // SAVING THE EDIT LOG
            $newVal = $this->get_data_for_obj($do->id);
            $this->createEditLog('created', 'delivery_orders', $do->id, $do->do_number,null, $newVal, $request->user->id,$request->current_year); // CREATE EDIT LOG
            
        
            // TRANSACTIONS COMMITTED
            DB::commit();
            
            return response()->json(['id'=>$do->id, 'do_number'=>$do->do_number], 200); // RETURNED RESPONSE

        } 
        // IF ANY ERROR WHILE SAVING
        catch (\Exception $e) {
            DB::rollback();
            return response()->json(['error' => $e->getMessage()], 500);
        }
    }

    public function update(Request $request, $pk){   
        $test = $this->is_edit_permitted($request->user->id, 'DeliveryOrder');
        $test_edit = $test->edit;
        $edit_valid_till_hours = $test->edit_valid_till;

        if( !$test_edit ){
            return response()->json(['Permission Denied For The Action'],401);
        }
        
        $do = DeliveryOrder::find($pk);

        if( $edit_valid_till_hours!=null ){
            $validity = Carbon::parse($do->created_at)->addHours($edit_valid_till_hours);
            if( !$validity->isFuture() ){
                return response()->json(['Permission Timed Out'],500);
            }
        }

        $validatedData = $request->validate([
            'branch' => 'required',
            'do_number' => 'required',
            'acc_party' => 'required'
        ], [
            'branch.required' => 'Required',
            'do_number.required' => 'Required',
            'acc_party.required' => 'required'
        ]);
          

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction();

            // FETCHING OLD VALUE FOR LOG
            $oldValue = $this->get_data_for_obj($pk);

            // $gr_check = $this->check_in_gr('id','gr_number', 'do_id', $pk);
                
            // if($gr_check['status']){
            //    $status = $gr_check['status'];
            //    return response()->json(['error'=>'Document in use', 'data'=>$status],500);
            // };


            // SAVING A NEW GR
            $do = DeliveryOrder::find($pk);

            $do->invoice_number = $request->invoice_number;
            $do->acc_party_id = $request->acc_party['id'];
            $do->from_date = $request->from_date;
            $do->to_date = $request->to_date;
            $do->is_locked = $request->is_locked;
            $do->order_no = $request->order_no;
            $do->po_no = $request->po_no;
            $do->acc_party_id = $request->acc_party['id'];
            $do->consignor_id = $request->consignor?$request->consignor['id']:null;
            $do->consignor_name = $request->consignor?$request->consignor['party_name']:null;
            $do->consignee_id = $request->consignee?$request->consignee['id']:null;
            $do->consignee_name = $request->consignee?$request->consignee['party_name']:null;
            $do->invc_val = $request->invc_val;
            $do->time_alloted = $request->time_alloted;

            $do->updated_by = $request->user->id;
            $do->save();

            $do_descs_delete = DeliveryOrderDesc::where('do_id',$do->id)->delete();

            $do_descs = $request->do_descs;
            foreach ($do_descs as $desc) {
                if(!$desc['source'] || !$desc['unit'] )
                continue;
                
                $doDtls = new DeliveryOrderDesc();
                $doDtls->do_number = $do->do_number;
                $doDtls->do_id = $do->id;
                $doDtls->unit = $desc['unit'];
                $doDtls->material_id = $desc['material'] ? $desc['material']['id']:null;
                $doDtls->rate = $desc['rate'];
                $doDtls->hire_rate = $desc['hire_rate'];
                $doDtls->loading_exp = $desc['loading_exp'];
                $doDtls->qty = $desc['qty'];
                $doDtls->source_id = $desc['source']['id'];
                $doDtls->source_name = $desc['source']['station_name'];
                $doDtls->destination_id = $desc['destination'] ? $desc['destination']['id']: null;
                $doDtls->destination_name = $desc['destination'] ? $desc['destination']['station_name']: null;
                $doDtls->eway_no = $desc['eway_no'];
                $doDtls->eway_dt = $desc['eway_dt'];
                $doDtls->item_desc = $desc['item_desc'];
                $doDtls->fin_year = $request->current_year;
                $doDtls->save();
            }

            // SAVING THE EDIT LOG
            $newVal = $this->get_data_for_obj($do->id);
            $this->createEditLog('updated', 'delivery_orders', $do->id, $do->do_number,$oldValue, $newVal, $request->user->id,$request->current_year); // CREATE EDIT LOG
            

            // TRANSACTIONS COMMITTED
            DB::commit();

            
            return response()->json("Success", 200); // RETURNED RESPONSE

        } catch (\Exception $e) {
            DB::rollback();
            return response()->json(['error' => $e->getMessage()], 500);
        }
    }

    public function destroy(Request $request, $pk){
        $test = $this->is_delete_permitted($request->user->id, 'DeliveryOrder');
        $test_delete = $test->delete;
        $delete_valid_till_hours = $test->delete_valid_till;

        if( !$test_delete ){
            return response()->json(['Permission Denied For The Action'],401);
        }

        // FETCHING OLD VALUE FOR LOG
        $oldValue = $this->get_data_for_obj($pk);

        
    try {
        // TRANSACTION BEGINS
        DB::beginTransaction();
    

        $do = DeliveryOrder::find($pk);

        // $gr_check = $this->check_in_gr('id','gr_number', 'do_id', $pk);
                
        //     if($gr_check['status']){
        //        $status = $gr_check['status'];
        //        return response()->json(['error'=>'Document in use', 'data'=>$status],500);
        //     };
        
        if( $delete_valid_till_hours!=null ){
            $validity = Carbon::parse($do->created_at)->addHours($delete_valid_till_hours);
            if( !$validity->isFuture() ){
                return response()->json(['Permission Timed Out'],500);
            }
        }
          
        $do_descs = DeliveryOrderDesc::where('do_id',$do->id)->delete();

        //FIND AND DELETE THE BRANCH
        $do->delete();
        // COMMISSIONS WILL BE CASCADED AUTOMATICALLY
        
        $this->createEditLog('deleted', 'delivery_orders', $do->id, $do->do_number, $oldValue, null, $request->user->id,$request->current_year); // CREATE EDIT LOG

        // TRANSACTIONS COMMITTED
            DB::commit();
            return response()->json("Success", 200); // RETURNED RESPONSE

        } catch (\Exception $e) {
            DB::rollback();
            return response()->json(['error' => $e->getMessage()], 500);
        }
    }


    public function autocomplete(Request $request){
        // Initialize the query builder
        $query = DeliveryOrder::select('id', 'do_number');

        $user_branch_ids = $request->user_branch_ids;
        if ( $user_branch_ids ) {
            $query->whereIn('branch_id', $user_branch_ids );
        }

       // Apply any additional filters based on other columns as needed
        if ($request->has('filterby')) {
            $orWhereFilters = isset($request->filterby['orWhere']) ? $request->filterby['orWhere'] : [];
            $whereFilters = isset($request->filterby['where']) ? $request->filterby['where'] : [];
            $whereInFilters = isset($request->filterby['whereIn']) ? $request->filterby['whereIn'] : [];

        // Loop through the whereFilters array
        foreach ($whereFilters as $whereFilter) {
            // Apply the filter to the query
            $query->where($whereFilter['key'], $whereFilter['value']);
        }

        // Loop through the orWhereFilters array
        // Loop through the orWhereFilters array
        $query->where(function($query) use ($orWhereFilters) {
            foreach ($orWhereFilters as $orWhereFilter) {
                // Apply the filter to the query
                $query->orWhere($orWhereFilter['key'], $orWhereFilter['value']);
            }
        });

        foreach ($whereInFilters as $whereInFilter) {
            if(isset($whereInFilter['key']) && isset($whereInFilter['values'])){
                $query->whereIn($whereInFilter['key'], $whereInFilter['values']);
            }
        }  
        
        }
        
        // Check if searchVariable is provided and apply the filter
        if ($request->has('searchVariable')) {
            $searchVariable = $request->searchVariable;
            $query->where('do_number', 'like', '%'. $searchVariable . '%');
        }

        // Limit the number of results
        $limit = $request->limit ? $request->limit : 10;

        $query->take($limit)->orderBy('id', 'desc');

        // Retrieve the filtered 
        $dos = $query->get();
        return response(['dos' => $dos, 'draw' => $request->draw]);
    }

    public function autocomplete_do_for_gr(Request $request) {
        // Initialize the query builder
        $query = DeliveryOrder::select('id', 'do_number')->where('is_locked',0);

        // I only want do_numbers that are not already in the book_grs table
        // $query->whereDoesntHave('book_gr');
        // $query->where(function ($query) use ($request) {
        //     $query->orWhereHas('book_gr', function ($query) {
        //         // Check balance condition here
        //         $query->whereRaw('(
        //             SELECT SUM(delivery_order_descs.qty) FROM delivery_order_descs 
        //             WHERE do_id = delivery_orders.id 
        //         ) > (
        //             SELECT SUM(CASE 
        //                     WHEN gr_descs.chg_unit = "PKG" THEN gr_descs.pkg
        //                     WHEN gr_descs.chg_unit = "KG" THEN gr_descs.material_weight 
        //                     WHEN gr_descs.chg_unit = "MT" THEN gr_descs.material_weight
        //                 END) FROM gr_descs WHERE material_id IN (
        //                 SELECT material_id FROM delivery_order_descs WHERE do_id = delivery_orders.id
        //             )
        //         )');
        //     });
            
            // Check if searchVariable is provided and apply the filter
                $searchVariable = $request->searchVariable ?? '';
                $query->where('do_number', 'like', '%' . $searchVariable . '%');
        // });

        // Limit the number of results
        $query->take($request->limit)->orderBy('id', 'desc');
    
        // Retrieve the filtered records
        $dos = $query->get();

    
        return response(['dos' => $dos]);
    }

    public function get_do_report(Request $request){
        $length = $request->length;
        $column = $request->column;
        $dir = $request->dir;
        $searchValue = $request->search;

        $report_data = $request->report_data;
                                                                                                                 
        $query = DeliveryOrder::with([
                        'branch:id,branch_name,branch_code', 'consignor:id,party_name', 'consignee:id,party_name',
                        'acc_party.party:id,account_id',
                        'do_desc',
                        'book_gr:id,gr_number,do_id',
                        'do_desc.source:id,station_name', 'do_desc.destination:id,station_name', 'do_desc.material:id,material_name'
        ]);

        $query->withCount(['book_gr as book_gr_count' => function($q) {
            $q->whereDoesntHave('bill_desc'); // Ensure bill_desc relationship doesn't exist
        }]);


                        if( $request->user->party_id ){
                            $prt = Party::select('account_id')->find($request->user->party_id);
                            
                            $query->where(function($query) use ($prt){
                                $query->where('acc_party_id', $prt->account_id)->where('acc_party_id','!=', null);
                            });
                        }
                        
        
        if(isset($report_data['branch']) ){
            $query->where('branch_id', $report_data['branch']['id']);
        }

        if(isset($report_data['do_num_from']) ){
            $query->where('do_number',$report_data['do_num_from']['do_number']);
        }


        if(isset($report_data['from_date']) && isset($report_data['to_date'])){
            $query->where('from_date','<=',$report_data['from_date'])->where('to_date','>=',$report_data['to_date']);
        }

        if(isset($report_data['consignor']) ){
            $query->where('consignor_id', $report_data['consignor']['id']);
        }

        if(isset($report_data['consignee']) ){
            $query->where('consignee_id', $report_data['consignee']['id']);
        }


        if (isset($report_data['source'])) {
            $query->whereHas('do_desc', function ($query) use ($report_data) {
                $query->where('source_id', $report_data['source']['id']);
            });
        }

        if (isset($report_data['destination'])) {
            $query->whereHas('do_desc', function ($query) use ($report_data) {
                $query->where('destination_id', $report_data['destination']['id']);
            });
        }

        if ($searchValue) {
            $query->where(function ($query) use ($searchValue) {
                $query->where('do_number', 'like','%'. $searchValue . '%')
                ->orWhereHas('acc_party', function ($query) use ($searchValue) {
                    $query->where('account_name', 'like', '%' . $searchValue . '%');
                });
            });
        }

        

        if ($length) {
            $data = $query->orderBy($column, $dir)->paginate($length);
        } else {
            $data = $query->orderBy($column, $dir)->paginate( $query->count() );
        }

          
        return response()->json(['data' => $data, 'draw' => $request->draw]);
    }

    public function get_do_detail_for_gr(Request $request, $pk) {
    // FETCH AND SEND
    $do = DeliveryOrder::with([
        'branch:id,branch_name,branch_code',
        'consignor:id,party_name,account_id',
        'consignee:id,party_name,account_id',
        'acc_party.party:id,account_id',
        'do_desc', 
        'do_desc.source:id,station_name',
        'do_desc.destination:id,station_name',
        'do_desc.material:id,material_name',
        'book_gr:id,do_id,gr_number'
    ])
    ->where('do_number',$pk)->first();

    if(!$do){
        return response()->json(['do' => null]);
    }

    // if ($do->book_gr) {
    //     // Transform do_desc here
    //     $transformedDoDesc = $do->do_desc->map(function($doDesc) {
            
    //         $balance = $this->find_remaining_balance_for_do_desc($doDesc->id);

    //         // Compare qty with the sum of pkg
    //         if ( $balance !== null) {
    //             if ($doDesc->qty > $balance['used_qty']) {
    //                 // Adjust qty if needed
    //                 $doDesc->qty = $doDesc->qty - $balance['used_qty'];
    //             } else {
    //                 // Set qty to 0 if pkg is greater or equal
    //                 $doDesc->qty = 0;
    //             }
    //         }
    //         return $doDesc;
    //     });
    // } else {
        // If no book_gr exists, keep original do_desc
        $transformedDoDesc = $do->do_desc;
   // }


    // You can attach the transformed do_desc back to the $do object if needed
    $do->setRelation('do_desc', $transformedDoDesc);

    return response()->json(['do' => $do]);
}

protected function find_remaining_balance_for_do_desc($do_desc_id){
    // Find the delivery order description
    $do_desc = DeliveryOrderDesc::find($do_desc_id);

    // Check if the delivery order description exists
    if ($do_desc) {
        // Find the related delivery order
        $book_gr_ids = BookGr::where('do_id', $do_desc['do_id'])->pluck('id');

        $sum_field = 'material_weight';
        if($do_desc->unit=='PKG'){
            $sum_field = 'pkg';
        }
        elseif($do_desc->unit=='KG' || $do_desc->unit=='MT' || $do_desc->unit=='Quintal'){
            $sum_field = 'material_weight';
        }

        // Find the sum of pkg belonging to book_gr relationships of do_desc
        $sum_pkg = GrDesc::whereIn('gr_id', $book_gr_ids)
                        ->where('material_id', $do_desc->material_id)
                        ->sum($sum_field);

        return ['used_qty'=>$sum_pkg, 'remaining_qty'=>$do_desc->qty - $sum_pkg, 'total_qty'=>$do_desc->qty]  ;
    }

}



    public function check_unique(Request $request){
        $key = $request->key;
        $value = $request->value;

        if(!$value){
            return response()->json(['exists' => false]);
        }

        $exists = DeliveryOrder::where($key, $value)->exists();
        if($exists){
            return response()->json(['errors' => [ $key  => ['Already Exists !']]],422);
        }

        return response()->json(['exists' => false]);
    }

    public function get_recent(Request $request) {

        $dosQuery = DeliveryOrder::select('id', 'do_number')->where('fin_year',$request->current_year);
         
        if ( !empty($request->user_branch_ids) ) {
            $dosQuery->whereIn('branch_id', $request->user_branch_ids );
        }

        
        $dos = $dosQuery->orderBy('id', 'desc')->take(10)->get();
        
        return response()->json(['dos' => $dos]);
    }


    public function get_next_do_number(Request $request){
        $type=$request->type;
        $prefix=$request->prefix;
        $branch_id = $request->branch_id;
        $branch_code = $request->branch_code;

        if($type=='SYSTEM'){
            $max = DeliveryOrder::where('branch_id',$branch_id)->max('max_number');
            $do_number = $branch_code.'-'.sprintf('%06d', $max+1);
            
            return response()->json(['do_number' => strtoupper($do_number), 'max_number' => $max+1]);
        }
    }



    public function get_gr_report(Request $request){
        $length = $request->length;
        $column = $request->column;
        $dir = $request->dir;
        $searchValue = $request->search;

        $report_data = $request->report_data;
                                                                                                                 
        $query = DeliveryOrder::with(
                            [
                                'branch:id,branch_name,branch_code', 'consignor:id,party_name', 'consignee:id,party_name',
                    'acc_party.party:id,account_id',
                    'do_desc',
                    'do_desc.source:id,station_name', 'do_desc.destination:id,station_name', 'do_desc.material:id,material_name'
                            ]
                        )->where('fin_year',$request->current_year);

        
        if(isset($report_data['branch']) ){
            $query->where('branch_id', $report_data['branch']['id']);
        }

        if(isset($report_data['do_num_from']) && isset($report_data['do_num_to'])){
            $query->where('do_number','>=',$report_data['do_num_from']['do_number'])->where('do_number','<=',$report_data['do_num_to']['do_number']);
        }


        if(isset($report_data['consignor']) ){
            $query->where('consignor_id', $report_data['consignor']['id']);
        }

        if(isset($report_data['consignee']) ){
            $query->where('consignee_id', $report_data['consignee']['id']);
        }

        if(isset($report_data['acc_party']) ){
            $query->where('acc_party_id', $report_data['acc_party']['id']);
        }

        if ($searchValue) {
            $query->where(function ($query) use ($searchValue) {
                $query->where('do_number', 'like', $searchValue . '%');
            });
        }

        if ($length) {
            $data = $query->orderBy($column, $dir)->paginate($length);
        } else {
            $data = $query->orderBy($column, $dir)->paginate( $query->count() );
        }
          
        return response()->json(['data' => $data, 'draw' => $request->draw]);
    }


    protected function get_data_for_obj($pk){
        // FETCH AND SEND
        $do = DeliveryOrder::with([
            'branch:id,branch_name,branch_code', 'consignor:id,party_name', 'consignee:id,party_name',
                    'acc_party.party:id,account_id',
                    'do_desc',
                    'do_desc.source:id,station_name', 'do_desc.destination:id,station_name', 'do_desc.material:id,material_name'
            ])->find($pk);

        return ['do'=>$do];
    }

    protected function get_max_number_from_do_number($input){
        // Extract the integer value using regular expression
        $pattern = '/[-\/](\d+)$/';
        preg_match($pattern, $input, $matches);
        
        if(empty($matches)){
            return intval($input);
        }

        $integerValue = intval(end($matches));
        return $integerValue;
    }


}
