<?php

namespace App\Http\Controllers\FleetOwner;

use App\Http\Controllers\Controller;
use App\Http\Traits\CommonTrait;
use App\Http\Traits\EditLogTrait;
use App\Http\Traits\PermissionsTrait;
use App\Models\Booking\BookGr;
use App\Models\FleetOwner\DieselBillDesc;
use App\Models\FleetOwner\DieselReceipt;
use App\Models\FleetOwner\DriverSettlementDesc;
use App\Models\FleetOwner\LoadingSlip;
use App\Models\GoodsMove\Challan;
use App\Models\GoodsMove\ChallanDesc;
use App\Models\Master\FuelAllocationMaster;
use Carbon\Carbon;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;

class DieselReceiptController extends Controller
{
    use PermissionsTrait;
    use EditLogTrait;
    use CommonTrait;

    public function index(Request $request){
        $test = $this->is_read_permitted($request->user->id, 'DieselReceipt');
        if(!$test){
            return response()->json(['Permission Denied For The Action'],401);
        }

        //FETCH AND SEND ALL
        $diesel_receipts = DieselReceipt::with([
                    'branch:id,branch_name,branch_code',
                    'vehicle:id,vehicle_number',
                    'dr_on_account:id,account_name',
                    'driver_on_account:id,account_name',
                    'filling_station:id,account_name'
                    ])->where('fin_year',$request->current_year)->get();
        return response()->json($diesel_receipts);
    }

    public function show(Request $request, $pk){
        // FETCH AND SEND
        $dr = DieselReceipt::with([
                        'branch:id,branch_name,branch_code',
                        'vehicle:id,vehicle_number,self_market',
                        'dr_on_account:id,account_name',
                        'driver_on_account:id,account_name',
                        'filling_station:id,account_name'
                    ])->find($pk);


        $challan = null;

        if($dr->against=='GR'){
            $challan = ChallanDesc::select('challan_id','challan_number')->where('gr_id',$dr->booking_id)->first();
        }

        return response()->json(['diesel_receipt' => $dr, 'challan'=>$challan]);
    }

    public function store(Request $request){

        $test = $this->is_new_permitted($request->user->id, 'DieselReceipt');
        if(!$test){
            return response()->json(['Permission Denied For The Action'],401);
        }
        
        $validatedData = $request->validate([
            'branch' => 'required',
            'dr_number' => 'required|unique:diesel_receipts,dr_number',
            'vehicle' => 'required',
            'dr_on_account' => 'required',
            'filling_station' => 'required',
        ], [
            'branch.required' => 'Required',
            'dr_number.required' => 'Required',
            'vehicle.required' => 'Required',
            'dr_on_account.required' => 'Required',
            'filling_station.required' => 'Required',
        ]);

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction();  

            $max_number = $this->get_max_number_from_dr_number($request->dr_number);      
            
            $send_exceed=null;
            if($request->against=='GR' && $request->booking_number){
                $balance = $this->find_balance_fuel($request->against, $request->booking_number['id']);
                if($balance['message']!="Route Fuel Not Found"){
                    if($balance['balance'] - $request->qty<0){
                        $send_exceed = ['message'=>'Diesel Receipt Saved But Fuel Exceeded The Route Qty'];
                        //return response()->json(['message'=>'No Balance Remaining For Fuel Allotment', 'remaining'=>abs($balance['balance'] - $request->qty)], 500);
                    }
                }
                 
            }

            
            // SAVING A NEW DR
            $dr = new DieselReceipt();
            $dr->branch_id = $request->branch['id'];
            $dr->max_number = $max_number;
            $dr->dr_number = $request->dr_number;
            $dr->against = $request->against;


            $dr->vehicle_id = $request->vehicle['id'];
            $dr->vehicle_number = $request->vehicle['vehicle_number'];
            $dr->booking_id = $request->booking_number ? $request->booking_number['id']:null;
            $dr->booking_number =  $request->booking_number ?  $request->booking_number['against_number'] :null;
            $dr->bk_date = $request->bk_date;
            $dr->booking_date = $request->booking_date;
            $dr->bill_number = $request->bill_number;
            $dr->dr_on_account = $request->dr_on_account['id'];
            $dr->filling_station = $request->filling_station['id'];
            $dr->slip_number = $request->slip_number;
            $dr->fuel_type = $request->fuel_type;
            $dr->qty = $request->qty;
            $dr->rate = $request->rate;
            $dr->amount = $request->amount;
            $dr->remarks = $request->remarks;
            $dr->reading = $request->reading;
            $dr->cash_amount = $request->cash_amount;
            $dr->driver_on_account = $request->driver_on_account? $request->driver_on_account['id']:null;

            $dr->fin_year = $request->current_year; 

            $dr->created_by = $request->user->id;

            // Save the model to the database
            $dr->save();
            
            
            // RECORD THE TRANSACTION IN ACCOUNTS
            $this->record_transaction_in_accounts($request, $dr);
            // TRANSACTION RECORDED IN ACCOUNTS

            // SAVING THE EDIT LOG
            $newVal = $this->get_data_for_obj($dr->id);
            $this->createEditLog('created', 'diesel_receipts', $dr->id, $dr->dr_number,null, $newVal, $request->user->id,$request->current_year); // CREATE EDIT LOG
            
        
            // TRANSACTIONS COMMITTED
            DB::commit();

            
            return response()->json(['data'=>['id'=>$dr->id, 'dr_number'=>$dr->dr_number], 'send_exceed'=>$send_exceed], 200); // RETURNED RESPONSE

        } 
        // IF ANY ERROR WHILE SAVING
        catch (\Exception $e) {
            DB::rollback();
            return response()->json(['error' => $e->getMessage()], 500);
        }
    }

    public function update(Request $request, $pk){   
        $test = $this->is_edit_permitted($request->user->id, 'DieselReceipt');
        $test_edit = $test->edit;
        $edit_valid_till_hours = $test->edit_valid_till;

        if( !$test_edit ){
            return response()->json(['Permission Denied For The Action'],401);
        }
        
        $dr = DieselReceipt::find($pk);
        $check_in_diesel_bill = DieselBillDesc::select('db_number')->where('dr_id',$pk)->first();
        if($check_in_diesel_bill){
            return response()->json(['error'=>'Document in use', 'data'=>['message'=>'Stopped By Diesel Bill : '.$check_in_diesel_bill->db_number]],500);
        }

        if( $edit_valid_till_hours!=null ){
            $validity = Carbon::parse($dr->created_at)->addHours($edit_valid_till_hours);
            if( !$validity->isFuture() ){
                return response()->json(['Permission Timed Out'],500);
            }
        }

        $validatedData = $request->validate([
            'vehicle' => 'required',
            'dr_on_account' => 'required',
            'filling_station' => 'required',
        ], [
            'vehicle.required' => 'Required',
            'dr_on_account.required' => 'Required',
            'filling_station.required' => 'Required',
        ]);
          

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction();

            // FETCHING OLD VALUE FOR LOG
            $oldValue = $this->get_data_for_obj($pk);

            $send_exceed=null;
            if($request->against=='GR' && $request->booking_number){
                $balance = $this->find_balance_fuel($request->against, $request->booking_number['id']);
                if($balance['message']!="Route Fuel Not Found"){
                    if($balance['balance'] + $dr->qty - $request->qty<0){
                        $send_exceed = ['message'=>'Diesel Receipt Saved But Fuel Exceeded The Route Qty'];
                        //return response()->json(['message'=>'No Balance Remaining For Fuel Allotment'], 500);
                    }
                }
                 
            }

            if( $dr->against =='GR' && $dr->booking_number ){
                $driver_settlement_check = DriverSettlementDesc::where('against_id',$dr->booking_id)->where('against_number',$dr->booking_number)->first();
                if($driver_settlement_check){
                    return response(['error'=>'Document in use', 'data'=>['message'=>'Stopped By Driver Settlement : '.$driver_settlement_check->ds_number]],500);
                }
            }


            // SAVING A NEW DR
            $dr->against = $request->against;
            $dr->vehicle_id = $request->vehicle['id'];
            $dr->vehicle_number = $request->vehicle['vehicle_number'];
            $dr->booking_id = $request->booking_number ? $request->booking_number['id']:null;
            $dr->booking_number =  $request->booking_number ?  $request->booking_number['against_number'] :null;
            $dr->bk_date = $request->bk_date;
            $dr->booking_date = $request->booking_date;
            $dr->bill_number = $request->bill_number;
            $dr->dr_on_account = $request->dr_on_account['id'];
            $dr->filling_station = $request->filling_station['id'];
            $dr->slip_number = $request->slip_number;
            $dr->fuel_type = $request->fuel_type;
            $dr->qty = $request->qty;
            $dr->rate = $request->rate;
            $dr->amount = $request->amount;
            $dr->remarks = $request->remarks;
            $dr->reading = $request->reading;
            $dr->cash_amount = $request->cash_amount;
            $dr->driver_on_account = $request->driver_on_account? $request->driver_on_account['id']:null;

            $dr->updated_by = $request->user->id;

            $dr->save();
            // NEW LS SAVED

            // RECORD THE TRANSACTION IN ACCOUNTS
            $this->record_transaction_in_accounts($request, $dr);
            // TRANSACTION RECORDED IN ACCOUNTS
            
            
            // SAVING THE EDIT LOG
            $newVal = $this->get_data_for_obj($dr->id);
            $this->createEditLog('updated', 'diesel_receipts', $dr->id, $dr->dr_number,$oldValue, $newVal, $request->user->id,$request->current_year); // CREATE EDIT LOG
            

            // TRANSACTIONS COMMITTED
            DB::commit();

            
            return response()->json(['data'=>['id'=>$dr->id, 'dr_number'=>$dr->dr_number], 'send_exceed'=>$send_exceed], 200);

        } catch (\Exception $e) {
            DB::rollback();
            return response()->json(['error' => $e->getMessage()], 500);
        }
    }

    public function destroy(Request $request, $pk){
        $test = $this->is_delete_permitted($request->user->id, 'DieselReceipt');
        $test_delete = $test->delete;
        $delete_valid_till_hours = $test->delete_valid_till;

        if( !$test_delete ){
            return response()->json(['Permission Denied For The Action'],401);
        }

        // FETCHING OLD VALUE FOR LOG
        $oldValue = $this->get_data_for_obj($pk);
        
        $dr = DieselReceipt::find($pk);
        $check_in_diesel_bill = DieselBillDesc::select('db_number')->where('dr_id',$pk)->first();
        if($check_in_diesel_bill){
            return response()->json(['error'=>'Document in use', 'data'=>['message'=>'Stopped By Diesel Bill : '.$check_in_diesel_bill->db_number]],500);
        }

        if( $dr->against =='GR' && $dr->booking_number ){
            $driver_settlement_check = DriverSettlementDesc::where('against_id',$dr->booking_id)->where('against_number',$dr->booking_number)->first();
            if($driver_settlement_check){
                return response(['error'=>'Document in use', 'data'=>['message'=>'Stopped By Driver Settlement : '.$driver_settlement_check->ds_number]],500);
            }
        }
        

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction();
            

        if( $delete_valid_till_hours!=null ){
            $validity = Carbon::parse($dr->created_at)->addHours($delete_valid_till_hours);
            if( !$validity->isFuture() ){
                return response()->json(['Permission Timed Out'],500);
            }
        }

        // DELETE THE TRANSACTION VOUCHER
        $t_data = [];
        $t_data['t_against'] = 'DIESEL RECEIPT';
        $t_data['t_against_no'] = $dr->dr_number;
        $t_data['fin_year'] = $request->current_year;
        $t_data['t_against_id'] = $dr->id;
        $this->delete_transaction_voucher($t_data);

        //FIND AND DELETE THE LS
        $dr->delete();

        $this->createEditLog('deleted', 'diesel_receipts', $dr->id, $dr->dr_number, $oldValue, null, $request->user->id,$request->current_year); // CREATE EDIT LOG

            // TRANSACTIONS COMMITTED
            DB::commit();
            return response()->json("Success", 200); // RETURNED RESPONSE

        } catch (\Exception $e) {
            DB::rollback();
            return response()->json(['error' => $e->getMessage()], 500);
        }
    }

    public function autocomplete(Request $request){
        // Initialize the query builder
        $query = DieselReceipt::select('id', 'dr_number')->where('fin_year',$request->current_year);

        $user_branch_ids = $request->user_branch_ids;
        if ( $user_branch_ids ) {
            $query->whereIn('branch_id', $user_branch_ids );
        }

       // Apply any additional filters based on other columns as needed
        if ($request->has('filterby')) {
            $orWhereFilters = isset($request->filterby['orWhere']) ? $request->filterby['orWhere'] : [];
            $whereFilters = isset($request->filterby['where']) ? $request->filterby['where'] : [];

        // Loop through the whereFilters array
        foreach ($whereFilters as $whereFilter) {
            // Apply the filter to the query
            $query->where($whereFilter['key'], $whereFilter['value']);
        }

        // Loop through the orWhereFilters array
        foreach ($orWhereFilters as $orWhereFilter) {
            // Apply the filter to the query
            $query->orWhere($orWhereFilter['key'], $orWhereFilter['value']);
        }
        
        }
        
        // Check if searchVariable is provided and apply the filter
        if ($request->has('searchVariable')) {
            $searchVariable = $request->searchVariable;
            $query->where('dr_number', 'like', '%'. $searchVariable . '%');
        }

        // Limit the number of results
        $query->take(25)->orderBy('max_number', 'asc');

        // Retrieve the filtered 
        $diesel_receipts = $query->get();
        return response(['diesel_receipts' => $diesel_receipts, 'draw' => $request->draw]);
    }

    public function autocomplete_dsl_page(Request $request) {
        // Initialize the query builder
        $query = DieselReceipt::select('id', 'dr_number', 'booking_date', 'vehicle_id')
                 ->with(['vehicle:id,vehicle_number'])
                 ->where('fin_year', $request->current_year);
    
        $user_branch_ids = $request->user_branch_ids;
        if ($user_branch_ids) {
            $query->whereIn('branch_id', $user_branch_ids);
        }
    
        if ($request->has('searchVariable')) {
            $searchVariable = $request->searchVariable;
            $query->where(function($subQuery) use ($searchVariable) {
                $subQuery->where('dr_number', 'like', '%' . $searchVariable . '%')
                         ->orWhereHas('vehicle', function($vehicleQuery) use ($searchVariable) {
                             $vehicleQuery->where('vehicle_number', 'like', '%' . $searchVariable . '%');
                         });
            });
        }
        
        // Limit the number of results
        $query->take(25)->orderBy('max_number', 'asc');
    
        // Retrieve the filtered results
        $diesel_receipts = $query->get();
    
        // Format the results to include vehicle_number at the root level
        $formatted_results = $diesel_receipts->map(function($receipt) {
            return [
                'id' => $receipt->id,
                'dr_number' => $receipt->dr_number,
                'booking_date' => $receipt->booking_date,
                'vehicle_number' => optional($receipt->vehicle)->vehicle_number,
            ];
        });
    
        return response(['diesel_receipts' => $formatted_results, 'draw' => $request->draw]);
    }
    

    public function against_for_diesel_receipt(Request $request){
        // Initialize the query builder
        $against = $request->against;
        $branch_id = $request->branch_id;
        $vehicle_id = $request->vehicle_id;
        $searchVariable = $request->searchVariable;

        if($against=='GR'){
            $query = BookGr::select('id', 'gr_number as against_number','max_number')->where('is_cancelled',false)
            ->when($vehicle_id, function($query) use ($vehicle_id) {
                return $query->where('vehicle_id', $vehicle_id);
            })
            ->where('gr_number', 'like', '%'. $searchVariable . '%')
            ->whereDoesntHave('d_settlement_desc');   
        }
        elseif($against == 'Loading Slip'){
            $query = LoadingSlip::select('id', 'ls_number as against_number','max_number')
            ->when($vehicle_id, function($query) use ($vehicle_id) {
                return $query->where('vehicle_id', $vehicle_id);
            })->where('ls_number', 'like', '%'. $searchVariable . '%')
            ->whereDoesntHave('d_settlement_desc');   
        }
        elseif($against == 'Challan'){
            $query = Challan::select('id', 'challan_number as against_number','max_number')
            ->when($vehicle_id, function($query) use ($vehicle_id) {
                return $query->where('vehicle_id', $vehicle_id);
            })->where('challan_number', 'like', '%'. $searchVariable . '%')
            ->whereDoesntHave('d_settlement_desc');   
        }
        // Check if searchVariable is provided and apply the filter
        // if ( $branch_id ) {
        //     $query->where('branch_id', $branch_id);
        // }

        // Limit the number of results
        $query->take(20)->orderBy('max_number', 'asc');

        // Retrieve the filtered 
        $againsts = $query->get();
        return response(['againsts' => $againsts]);
    }

    public function get_details_for_against(Request $request){
        $against = $request->against;
        $booking_id = $request->booking_number['id'];
        
        $account = null;
        $vehicle = null;
        $party_account = null;
        $challan = null;

        if($against=='GR'){
            $query = BookGr::select('id', 'gr_number as against_number', 'gr_date as date','driver_id', 'vehicle_id', 'source_id','destination_id')
                    ->with(['driver:id,emp_name,account_id','driver.account:id,account_name','vehicle:id,vehicle_number,self_market,owner_id','challan_desc:id,gr_id,gr_number,challan_number,challan_id','vehicle.party:id,party_name,account_id','vehicle.party.account:id,account_name'])
                    ->find( $booking_id ); 

            $account = $query->driver ? $query->driver->account : null;
            $vehicle = $query->vehicle;
            $party_account = $query->vehicle ? $query->vehicle->party ? $query->vehicle->party->account : null : null;
            $challan = $query->challan_desc;
        }
        elseif($against == 'Loading Slip'){
            $query = LoadingSlip::select('id', 'ls_number as against_number', 'ls_date as date','driver_id', 'vehicle_id', 'source_id','destination_id')
                    ->with(['driver:id,emp_name,account_id','vehicle:id,vehicle_number,self_market,owner_id','vehicle.party:id,party_name,account_id','vehicle.party.account:id,account_name']) 
                    ->find( $booking_id );
            $account = $query->driver; 
            $vehicle = $query->vehicle;
            $party_account = $query->vehicle ? $query->vehicle->party ? $query->vehicle->party->account : null : null;
        }
        elseif($against == 'Challan'){
            $query = Challan::select('id', 'challan_number as against_number', 'challan_date as date','driver_id', 'vehicle_id', 'source_id','destination_id')
                    ->with(['driver:id,emp_name,account_id','vehicle:id,vehicle_number,self_market,owner_id','vehicle.party:id,party_name,account_id','vehicle.party.account:id,account_name'])
                    ->find( $booking_id ); 
            
            $account = $query->driver; 
            $vehicle = $query->vehicle;
            $party_account = $query->vehicle ? $query->vehicle->party ? $query->vehicle->party->account : null : null;
        }


        $balance = 0;

        $fuel_allocation = FuelAllocationMaster::where('source_id',$query->source_id)->where('destination_id',$query->destination_id)->first();

        if($fuel_allocation){
            $route_fuel = $fuel_allocation->fuel_qty;

            $diesel_fuel = DieselReceipt::where('against',$against)->where('booking_id',$booking_id)->sum('qty');
            $balance = $route_fuel - $diesel_fuel;            
        }
        else{
            $balance = 'Undefined';
        }


        return response()->json(['driver_account'=>$account, 'data'=>$query, 'vehicle'=>$vehicle, 'party_account'=>$party_account, 'balance_fuel'=>$balance, 'challan'=>$challan]);

    }

    public function store_multiple_diesel_receipt(Request $request){
        $test = $this->is_new_permitted($request->user->id, 'DieselReceipt');
        if(!$test){
            return response()->json(['Permission Denied For The Action'],401);
        }

        $diesel_receipts = $request->diesel_receipts;

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction(); 
        
        foreach ($diesel_receipts as $index => $pl) {
            if( !$pl['branch'] || !$pl['vehicle'] || !$pl['dr_on_account'] || !$pl['filling_station'] || !$pl['booking_date'] || !$pl['amount'] )
                continue;
            

            $next_dr_number = $this->number_generator_for_multi($pl['branch']);
            $dr_number = $next_dr_number['dr_number'];
            $max_number = $next_dr_number['max_number'];

            $dr = new DieselReceipt();
            $dr->branch_id = $pl['branch']['id'];
            $dr->max_number = $max_number;
            $dr->dr_number = $dr_number ;
            $dr->against = $pl['against'];
            $dr->vehicle_id = $pl['vehicle']['id'];
            $dr->vehicle_number = $pl['vehicle']['vehicle_number'];
            $dr->booking_id = $pl['booking_number']? $pl['booking_number']['id'] :null;
            $dr->booking_number = $pl['booking_number']? $pl['booking_number']['against_number'] :null;
            $dr->booking_date = $pl['booking_date'];
            $dr->dr_on_account = $pl['dr_on_account']['id'];
            $dr->filling_station = $pl['filling_station']['id'];
            $dr->bill_number = $pl['bill_number'];
            $dr->slip_number = $pl['slip_number'];
            $dr->reading = $pl['reading'];
            $dr->fuel_type = $pl['fuel_type'];
            $dr->qty = $pl['qty'];
            $dr->rate = $pl['rate'];
            $dr->amount = $pl['amount'];
            $dr->driver_on_account = $pl['driver_on_account']? $pl['driver_on_account']['id'] :null;
            $dr->cash_amount = $pl['cash_amount'];
            $dr->remarks = $pl['remarks'];

            $dr->fin_year = $request->current_year;
            $dr->created_by = $request->user->id; 
            $dr->save();

            $pl['current_year'] = $request->current_year;
            
           
            // RECORD THE TRANSACTION IN ACCOUNTS
            $this->record_transaction_in_accounts($pl, $dr);

            
         

            // SAVING THE EDIT LOG
            $newVal = $this->get_data_for_obj($dr->id);
            $this->createEditLog('created', 'diesel_receipts', $dr->id, $dr->dr_number,null, $newVal, $request->user->id,$request->current_year); // CREATE EDIT LOG
        }


             DB::commit();
            
            return response()->json(['success' => 'Diesel List Created Successfully']);

        } catch (\Exception $e) {
            DB::rollback();
            return response()->json(['errors' => ['diesel_receipt' => [$e->getMessage()]]], 422);
        } 
    
    }

    public function show_multiple_diesel_receipt(Request $request){
        $query = DieselReceipt::with(['branch:id,branch_name,branch_code','vehicle:id,vehicle_number,vehicle_type,self_market','dr_on_account:id,account_name','filling_station:id,account_name','driver_on_account:id,account_name'])->where('fin_year',$request->current_year);

        if($request->filling_station){
            $query->where('filling_station',$request->filling_station['id']);
        }

        if($request->vehicle){
            $query->where('vehicle_id',$request->vehicle['id']);
        }

        if($request->from_date){
            $query->where('booking_date','>=',$request->from_date);
        }

        if($request->to_date){
            $query->where('booking_date','<=',$request->to_date);
        }

        $result = $query->get();

        // Map booking_number
        $result->map(function($item){
            $item['booking_number'] = $item->booking_id? ['id'=>$item->booking_id, 'against_number'=>$item->booking_number] : null;
            return $item;
        });

        return response()->json(['diesel_receipt'=>$result]);
    }

    public function update_multi_diesel_receipt(Request $request){
        $test = $this->is_edit_permitted($request->user->id, 'DieselReceipt');
        $test_edit = $test->edit;
        $edit_valid_till_hours = $test->edit_valid_till;


        if( !$test_edit ){
            return response()->json(['Permission Denied For The Action'],401);
        }

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction();  

            $newly_added_diesel_receipts = $request->newly_added_diesel_receipts;

            foreach ($newly_added_diesel_receipts as $nadr) {
                
                $next_dr_number = $this->number_generator_for_multi($nadr['branch']);
                $dr_number = $next_dr_number['dr_number'];
                $max_number = $next_dr_number['max_number'];

                $dr = new DieselReceipt();
                $dr->branch_id = $nadr['branch']['id'];
                $dr->max_number = $max_number;
                $dr->dr_number = $dr_number;
                $dr->against = $nadr['against'];
                $dr->vehicle_id = $nadr['vehicle']['id'];
                $dr->vehicle_number = $nadr['vehicle']['vehicle_number'];
                $dr->booking_id = $nadr['booking_number']? $nadr['booking_number']['id'] :null;
                $dr->booking_number = $nadr['booking_number']? $nadr['booking_number']['against_number'] :null;
                $dr->booking_date = $nadr['booking_date'];
                $dr->dr_on_account = $nadr['dr_on_account']['id'];
                $dr->filling_station = $nadr['filling_station']['id'];
                $dr->bill_number = $nadr['bill_number'];
                $dr->slip_number = $nadr['slip_number'];
                $dr->fuel_type = $nadr['fuel_type'];
                $dr->qty = $nadr['qty'];
                $dr->rate = $nadr['rate'];
                $dr->amount = $nadr['amount'];
                $dr->remarks = $nadr['remarks'];
                $dr->reading = $nadr['reading'];
                $dr->driver_on_account = $nadr['driver_on_account']? $nadr['driver_on_account']['id'] :null;
                $dr->cash_amount = $nadr['cash_amount'];

                $dr->fin_year = $request->current_year;
                $dr->created_by = $request->user->id; 
                $dr->save();

                $nadr['current_year'] = $request->current_year;

                // RECORD THE TRANSACTION IN ACCOUNTS
                $this->record_transaction_in_accounts($nadr, $dr);

                // SAVING THE EDIT LOG
                $newVal = $this->get_data_for_obj($dr->id);
                $this->createEditLog('created', 'diesel_receipts', $dr->id, $dr->dr_number,null, $newVal, $request->user->id,$request->current_year); // CREATE EDIT LOG
            }



            $updated_diesel_receipts = $request->updated_diesel_receipts;

            foreach ($updated_diesel_receipts as $udr) {
                $dr = DieselReceipt::find($udr['id']);

                // FETCHING OLD VALUE FOR LOG
                $oldValue = $this->get_data_for_obj($udr['id']);

                // SAVING A NEW DR
                $dr->dr_number = $udr['dr_number'];
                $dr->against = $udr['against'];
                $dr->vehicle_id = $udr['vehicle']['id'];
                $dr->vehicle_number = $udr['vehicle']['vehicle_number'];
                $dr->booking_id = $udr['booking_number'] ? $udr['booking_number']['id'] :null;
                $dr->booking_number = $udr['booking_number'] ? $udr['booking_number']['against_number'] :null;
                $dr->booking_date = $udr['booking_date'];
                $dr->bill_number = $udr['bill_number'];
                $dr->dr_on_account = $udr['dr_on_account']['id'];
                $dr->filling_station = $udr['filling_station']['id'];
                $dr->bill_number = $udr['bill_number'];
                $dr->slip_number = $udr['slip_number'];
                $dr->fuel_type = $udr['fuel_type'];
                $dr->qty = $udr['qty'];
                $dr->rate = $udr['rate'];
                $dr->amount = $udr['amount'];
                $dr->remarks = $udr['remarks'];
                $dr->reading = $udr['reading'];
                $dr->driver_on_account = $udr['driver_on_account']? $udr['driver_on_account']['id'] :null;
                $dr->cash_amount = $udr['cash_amount'];

                $dr->updated_by = $request->user->id;

                $dr->save();
                // NEW LS SAVED

                $udr['current_year'] = $request->current_year;

                // RECORD THE TRANSACTION IN ACCOUNTS
                $this->record_transaction_in_accounts($udr, $dr);
                // TRANSACTION RECORDED IN ACCOUNTS

                // SAVING THE EDIT LOG

                $newVal = $this->get_data_for_obj($dr->id);
                $this->createEditLog('updated', 'diesel_receipts', $dr->id, $dr->dr_number,$oldValue, $newVal, $request->user->id,$request->current_year); // CREATE EDIT LOG
            }
            

            $deleted_diesel_receipts = $request->deleted_diesel_receipts;
            foreach ($deleted_diesel_receipts as $ddr) {

                $test = $this->is_delete_permitted($request->user->id, 'DieselReceipt');
                $test_delete = $test->delete;
                $delete_valid_till_hours = $test->delete_valid_till;

                if( !$test_delete ){
                    return response()->json(['Permission Denied For The Action'],401);
                }

                    $oldValue = $this->get_data_for_obj($ddr);
                    
                    $dr = DieselReceipt::find($ddr);

                    // DELETE THE TRANSACTION VOUCHER
                    $t_data = [];
                    $t_data['t_against'] = 'DIESEL RECEIPT';
                    $t_data['t_against_no'] = $dr->dr_number;
                    $t_data['fin_year'] = $request->current_year;
                    $t_data['t_against_id'] = $dr->id;
                    $this->delete_transaction_voucher($t_data);

                    //FIND AND DELETE THE LS
                    $dr->delete();

                    $this->createEditLog('deleted', 'diesel_receipts', $dr->id, $dr->dr_number, $oldValue, null, $request->user->id,$request->current_year); // CREATE EDIT LO=G
            }

            DB::commit();

            return response()->json(['success' => 'Diesel List Updated Successfully']);
        } catch (\Exception $e) {
            DB::rollback();
            return response()->json(['errors' => ['diesel_receipt' => [$e->getMessage()]]], 422);
        }

    }

    public function destroy_multi_diesel_receipt(Request $request){

        $test = $this->is_delete_permitted($request->user->id, 'DieselReceipt');
        $test_delete = $test->delete;
        $delete_valid_till_hours = $test->delete_valid_till;

        if( !$test_delete ){
            return response()->json(['Permission Denied For The Action'],401);
        }

        try {
            // TRANSACTION BEGINS
            DB::beginTransaction();  

            $deleted_diesel_receipts = $request->deleted_diesel_receipts;

            foreach ($deleted_diesel_receipts as $ddr) {

                    $oldValue = $this->get_data_for_obj($ddr);
                    
                    $dr = DieselReceipt::find($ddr);

                    if( $delete_valid_till_hours!=null ){
                        $validity = Carbon::parse($dr->created_at)->addHours($delete_valid_till_hours);
                        if( !$validity->isFuture() ){
                            return response()->json(['Permission Timed Out'],500);
                        }
                    }

                    // DELETE THE TRANSACTION VOUCHER
                    $t_data = [];
                    $t_data['t_against'] = 'DIESEL RECEIPT';
                    $t_data['t_against_no'] = $dr->dr_number;
                    $t_data['fin_year'] = $request->current_year;
                    $t_data['t_against_id'] = $dr->id;
                    $this->delete_transaction_voucher($t_data);

                    //FIND AND DELETE THE LS
                    $dr->delete();

                    $this->createEditLog('deleted', 'diesel_receipts', $dr->id, $dr->dr_number, $oldValue, null, $request->user->id,$request->current_year); // CREATE EDIT LOG
            }

            DB::commit();

            return response()->json(['success' => 'Diesel List Deleted Successfully']);
        } catch (\Exception $e) {
            DB::rollback();
            return response()->json(['errors' => ['diesel_receipt' => [$e->getMessage()]]], 422);
        }

    }


    protected function number_generator_for_multi($data){
            $branch_id = $data['id'];
            $branch_code = $data['branch_code'];
            
            $max = DieselReceipt::where('branch_id',$branch_id)->max('max_number');
            $dr_number = $branch_code.'-'.sprintf('%06d', $max+1);
            return ['dr_number' => strtoupper($dr_number), 'max_number' => $max+1];
    }

    public function datatable(Request $request){
        $test = $this->is_read_permitted($request->user->id, 'DieselReceipt');
        if(!$test){
            return response()->json(['Permission Denied For The Action'],401);
        }

        $length = $request->length;
        $column = $request->column;
        $dir = $request->dir;
        $searchValue = $request->search;

        $query = DieselReceipt::with([
                    'branch:id,branch_name,branch_code',
                    'vehicle:id,vehicle_number',
                    'dr_on_account:id,account_name',
                    'driver_on_account:id,account_name',
                    'filling_station:id,account_name'
                    ])->where('fin_year',$request->current_year) 
                ->orderBy($column, $dir);

        if ($searchValue) {
            $query->where(function ($query) use ($searchValue) {
                $query->where('dr_number', 'like', $searchValue . '%');
            });
        }

        $data = $query->paginate($length);

        return response()->json(['data' => $data, 'draw' => $request->draw]);
    }

    public function check_unique(Request $request){
        $key = $request->key;
        $value = $request->value;

        if(!$value){
            return response()->json(['exists' => false]);
        }

        $exists = DieselReceipt::where($key, $value)->exists();
        if($exists){
            return response()->json(['errors' => [ $key  => ['Already Exists !']]],422);
        }

        return response()->json(['exists' => false]);
    }

    public function get_recent(Request $request) {
        $DieselReceiptsQuery = DieselReceipt::select('id', 'dr_number')->where('fin_year',$request->current_year);
    
        if (!empty($request->user_branch_ids)) {
            $DieselReceiptsQuery->whereIn('branch_id', $request->user_branch_ids);
        }
    
        $DieselReceipts = $DieselReceiptsQuery->orderBy('id', 'desc')->take(10)->get();
    
        return response()->json(['diesel_receipts' => $DieselReceipts]);
    }
    


    public function get_next_dr_number(Request $request){
        $branch_id = $request->branch_id;
        $branch_code = $request->branch_code;
        
        $max = DieselReceipt::where('branch_id',$branch_id)->max('max_number');
        $dr_number = $branch_code.'-'.sprintf('%06d', $max+1);
        return response()->json(['dr_number' => strtoupper($dr_number), 'max_number' => $max+1]);
        
    }


    public function get_diesel_receipt_report(Request $request){
        $length = $request->length;
        $column = $request->column;
        $dir = $request->dir;
        $searchValue = $request->search;

        $report_data = $request->report_data;
                                                                                                                 
        $query = DieselReceipt::with(['branch:id,branch_name,branch_code','vehicle:id,vehicle_number','dr_on_account:id,account_name','driver_on_account:id,account_name','filling_station:id,account_name'])->where('fin_year',$request->current_year);

        
        if(isset($report_data['branch']) ){
            $query->where('branch_id', $report_data['branch']['id']);
        }

        if(isset($report_data['dr_num_from']) && isset($report_data['dr_num_to'])){
            $query->where('dr_number','>=',$report_data['dr_num_from']['dr_number'])->where('dr_number','<=',$report_data['dr_num_to']['dr_number']);
        }


        if(isset($report_data['from_date']) && isset($report_data['to_date'])){
            $query->where('booking_date','>=',$report_data['from_date'])->where('booking_date','<=',$report_data['to_date']);
        }

        if(isset($report_data['vehicle']) ){
            $query->where('vehicle_id', $report_data['vehicle']['id']);
        }

        if(isset($report_data['filling_station']) ){
            $query->where('filling_station', $report_data['filling_station']['id']);
        }
        if ( !empty($request->user_branch_ids) ) {
            $query->whereIn('branch_id', $request->user_branch_ids );
        }


        if ($searchValue) {
            $query->where(function ($query) use ($searchValue) {
                $query->where('dr_number', 'like','%'. $searchValue . '%');
            });
        }

        if ($length) {
            $data = $query->orderBy($column, $dir)->paginate($length);
        } else {
            $data = $query->orderBy($column, $dir)->paginate( $query->count() );
        }
          
        return response()->json(['data' => $data, 'draw' => $request->draw]);
    }


    protected function find_balance_fuel($against, $against_id){
        $route_fuel = 0;

        $diesel_fuel = DieselReceipt::where('against',$against)->where('booking_id',$against_id)->sum('qty');

        if($against=='GR'){
            $gr = BookGr::select('id','gr_number','gr_date','source_id','destination_id')->find($against_id);

            $fuel_allocation = FuelAllocationMaster::where('source_id',$gr->source_id)->where('destination_id',$gr->destination_id)->first();
        }

        if($against=='Loading Slip'){
            $ls = LoadingSlip::select('id','gr_number','gr_date','source_id','destination_id')->find($against_id);
            $fuel_allocation = FuelAllocationMaster::where('source_id',$ls->source_id)->where('destination_id',$ls->destination_id)->first();
        }


        if($fuel_allocation){
            $route_fuel = $fuel_allocation->fuel_qty;
        }

        if(!$route_fuel){
            return ['balance'=>0, 'message'=>'Route Fuel Not Found'];
        }

        $balance = $route_fuel - $diesel_fuel;

        return ['balance'=>$balance, 'message'=>'Success'];
    }

    protected function get_data_for_obj($pk){
        // FETCH AND SEND
        $dr = DieselReceipt::with([
                        'branch:id,branch_name,branch_code',
                        'vehicle:id,vehicle_number',
                        'dr_on_account:id,account_name',
                        'driver_on_account:id,account_name',
                        'filling_station:id,account_name'
            ])->find($pk);

        return ['diesel_receipt' => $dr];
    }

    protected function get_max_number_from_dr_number($input){
        // Extract the integer value using regular expression
        $pattern = '/[-\/](\d+)$/';
        preg_match($pattern, $input, $matches);
        
        if(empty($matches)){
            return intval($input);
        }

        $integerValue = intval(end($matches));
        return $integerValue;
    }

    protected function record_transaction_in_accounts($request, $dr){
        $t_data = [];
        $t_data['t_against'] = 'DIESEL RECEIPT';
        $t_data['t_against_id'] = $dr->id;
        $t_data['t_against_no'] = $dr->dr_number;
        
        
        $t_data['fin_year'] = $request['current_year'];


        $this->delete_transaction_voucher($t_data);
        
        $t_data['branch'] = $request['branch'];
        $t_data['voucher_date'] = $request['booking_date'];
        $t_data['voucher_time'] = null;
        $t_data['t_against_narration'] = $dr->remarks;
        $t_data['fin_year'] = $request['current_year'];

        $t_data['voucher_accounting_details'] = [];
        
        $t_data['cancellable'] = 1;
        if($dr->dr_on_account && $dr->filling_station && $request['amount']>0){
            $t_data['voucher_accounting_details']=[];
            // DEBIT PARTY ACCOUNT
            $t_data['voucher_accounting_details'][] = ['account_id'=>$dr->dr_on_account, 'nature'=>'dr' , 'dr_amount'=>$request['amount'], 'cr_amount'=>0 ];
            

            // CREDIT INCOME ACCOUNT
            $t_data['voucher_accounting_details'][] = ['account_id'=>$dr->filling_station, 'nature'=>'cr' , 'dr_amount'=>0, 'cr_amount'=>$request['amount'] ];

            
            //CHECK TOTAL DR AND CR
            $total_dr = 0;
            $total_cr = 0;

            foreach ($t_data['voucher_accounting_details'] as $vad) {
                $total_dr += $vad['dr_amount'];
                $total_cr += $vad['cr_amount'];
            }

            // Round both $total_dr and $total_cr to two decimal places
            $total_dr_rounded = round(floatval($total_dr), 2);
            $total_cr_rounded = round(floatval($total_cr), 2);

            // Check if they are not equal
            if ($total_dr_rounded != $total_cr_rounded) {
                // Output the values for debugging
                dd($total_dr_rounded, $total_cr_rounded);
                
                // Return response with an error message
                return response()->json(['message' => 'Error In Accounting Entries'], 422);
            }


            $t_data['total_dr_amount'] = $total_dr;
            $t_data['total_cr_amount'] = $total_cr; 

            $this->create_transaction_voucher( $t_data );
        }

        $t_data['cancellable'] = 1;
    
        if($dr->driver_on_account && $dr->filling_station && $request['cash_amount']>0){
            $t_data['voucher_accounting_details']=[];
            // DEBIT PARTY ACCOUNT
            $t_data['voucher_accounting_details'][] = ['account_id'=>$dr->driver_on_account, 'nature'=>'dr' , 'dr_amount'=>$request['cash_amount'], 'cr_amount'=>0 ];
            

            // CREDIT INCOME ACCOUNT
            $t_data['voucher_accounting_details'][] = ['account_id'=>$dr->filling_station, 'nature'=>'cr' , 'dr_amount'=>0, 'cr_amount'=>$request['cash_amount'] ];

            
            //CHECK TOTAL DR AND CR
            $total_dr = 0;
            $total_cr = 0;

            foreach ($t_data['voucher_accounting_details'] as $vad) {
                $total_dr += $vad['dr_amount'];
                $total_cr += $vad['cr_amount'];
            }

            if($total_dr != $total_cr){
                return response()->json(['message'=>'Error In Accounting Entries'], 422);
            }


            $t_data['total_dr_amount'] = $total_dr;
            $t_data['total_cr_amount'] = $total_cr; 

            $this->create_transaction_voucher( $t_data );
        }

    }
}
